const fs = require("fs");
const path = require("path");
const { walk } = require("walk");
const { promisify } = require("util");
const AdmZip = require("adm-zip");
const _ = require("lodash");

const readFileAsync = promisify(fs.readFile);
const readdirAsync = promisify(fs.readdir);
const resolve = (...p) => path.resolve(__dirname, ...p);

exports.buildZip = buildZip;

/**
 * Build a magno widget from givne UMD modules.
 *
 * @param {{ name: string; description: string; author: string; entry: string; modulePath: string }} meta the metadata.
 * @returns {Promise<AdmZip>}
 */
async function buildZip(meta) {
  const manifest = constructManifest(meta);

  const zip = new AdmZip();

  // read and replace constants.
  (await generateFromWidgetTemplate(manifest)).forEach(({ name, data }) => {
    zip.addFile(name, Buffer.from(data, 'utf-8'));
  });

  // add widget metadata.
  const metadata = JSON.stringify({
    widgetApi: [
      {
        name: manifest.widgetName,
      },
    ],
    widgetDescription: manifest.description,
    authorName: manifest.author,
    localFileBasePath: '',
    requires: manifest.requires,
  });
  zip.addFile('packageinfo.json', Buffer.from(metadata, 'utf-8'));

  // add files recursively.

  const walker = walk(manifest.modulePath);

  return new Promise((resolve, reject) => {
    walker.on("file", (base, stat, next) => {
      const fullPath = `${base}/${stat.name}`;
      fs.readFile(fullPath, (err, data) => {
        if (err) {
          return reject(err);
        }
        const relativePath = path.relative(manifest.modulePath, fullPath);
        zip.addFile(`dist/${relativePath}`, Buffer.from(data, 'utf-8'));
        next();
      });
    });
    walker.on("errors", (_, nodeStats) => {
      reject(nodeStats);
    });
    walker.on("end", () => resolve(zip));
  });
}

/**
 * Construct the formal manifest.
 * 
 * @param {{ name: string; description: string; author: string; entry: string; modulePath: string }} meta the metadata.
 */
function constructManifest(meta) {
  return {
    name: meta.name,
    description: meta.description,
    author: meta.author,
    modulePath: meta.modulePath,
    entry: meta.entry,
    widgetName: `${_.upperFirst(meta.name)}Widget`,
    props: meta.props || [],
    events: meta.events || [],
    actions: meta.actions || [],
    requires: meta.requires || [],
  };
}

/**
 * Generate code from widget template.
 */
async function generateFromWidgetTemplate(manifest) {
  return Promise.all(
    (await readdirAsync(resolve("./template"))).map(async (f) => {
      return {
        name: f.replace("{widget}", manifest.widgetName),
        data: _.template(
          (await readFileAsync(resolve("./template", f))).toString(),
          {
            interpolate: /i?d?\/\*\{([\s\S]+?)\}\*\//,
          }
        )({ manifest }),
      };
    })
  );
}
